package heaps;

import java.util.Collection;
import java.util.Comparator;
import java.util.NoSuchElementException;

public class MaxHeap<E> implements IBinaryHeap<E> {

    private Object[] heapArray;
    private int size;

    public MaxHeap() {
        heapArray = new Object[2];
        size = 0;
    }

    public MaxHeap(Collection<? extends E> values) {
        size = values.size();
        heapArray = new Object[size];
        int i = 0;
        for (Object value : values) {
            if (value == null) {
                throw new IllegalArgumentException("No null values permitted in the heap");
            }
            heapArray[i++] = value;
        }
        heapify();
    }

    private static int parent(int i) { return (i - 1) / 2; }
    private static int left(int i)   { return 2 * i + 1;   }
    private static int right(int i)  { return 2 * i + 2;   }

    private void grow() {
        Object[] newArray =  new Object[heapArray.length * 2];
        System.arraycopy(heapArray, 0, newArray, 0, size);
        heapArray = newArray;
    }

    private void shrink() {
        Object[] newArray =  new Object[heapArray.length / 2];
        System.arraycopy(heapArray, 0, newArray, 0, size);
        heapArray = newArray;
    }

    private void siftUp(int index) {
        while (index > 0) {
            int parent = parent(index);
            Comparable<? super E> key = (Comparable<? super E>) heapArray[index];
            Object parentKey = heapArray[parent];

            // Check for max heap property violation
            if (key.compareTo((E) parentKey) <= 0) {
                return; // if this is <= parent, then no more sifting to do
            } else { // otherwise, swap this with parent and keep moving up
                Object temp = heapArray[index];
                heapArray[index] = heapArray[parent];
                heapArray[parent] = temp;
                index = parent;
            }
        }
    }

    private void siftDown(int index) {
        E value = (E) heapArray[index];
        Comparable<? super E> key = (Comparable<? super E>) heapArray[index];

        while (left(index) < size) {
            // find the max among the node and its children
            E maxValue = value;
            int maxIndex = index;
            if (key.compareTo((E) heapArray[left(index)]) < 0) {
                maxIndex = left(index);
                maxValue = (E) heapArray[left(index)];
            }
            if (right(index) < size && key.compareTo((E) heapArray[right(index)]) < 0) {
                maxIndex = right(index);
                maxValue = (E) heapArray[right(index)];
            }

            // check for a violation of the max heap property
            if (key.compareTo(maxValue) == 0) {
                return; // if this parent is the largest value among its children, no problem
            } else {
                // Swap downwards
                Object temp = heapArray[index];
                heapArray[index] = heapArray[maxIndex];
                heapArray[maxIndex] = temp;

                // continue loop from the index of the node we swapped
                index = maxIndex;
            }
        }

    }


    @Override
    public boolean add(E e) {
        // resize if needed
        if (size == heapArray.length) {
            grow();
        }

        // Add the new value to the end of the array
        heapArray[size++] = e;

        // sift up from the index we added to
        siftUp(size - 1);
        return true;
    }

    public E poll() {
        // Save the max value from the root of the heap
        E maxValue = (E) heapArray[0];

        // Move the last item in the array into index 0
        E replaceValue = (E) heapArray[--size];
        if (size > 0) {
            heapArray[0] = replaceValue;

            // Sift down to restore max heap property
            siftDown(0);
        }
        return maxValue;
    }

    @Override
    public boolean remove(Object o) {
        int idx = indexOf(o);
        if (idx == -1) {
            return false;
        }
        E replaceValue = (E) heapArray[--size];
        if (size > 0) {
            heapArray[0] = replaceValue;

            // Sift down to restore max heap property
            siftDown(idx);
        }
        return true;

    }

    private int indexOf(Object o) {
        Comparable<? super E> key = (Comparable<? super E>) o;
        for (int i = 0; i < size; i++) {
            if (o.equals(heapArray[i])) {
                return i;
            }
        }
        return -1;
    }

    @Override
    public E peek() {
        if (size == 0) {
            throw new NoSuchElementException();
        }
        return (E) heapArray[0];
    }

    @Override
    public boolean contains(Object o) {
        return indexOf(o) != -1;
    }

    private void heapify() {
    }
}
